From 90b07ecb71c8440a8d870b90fc2bea89ed71e54d Mon Sep 17 00:00:00 2001
From: Jarmo Torvinen <jarmo.torvinen@jutel.fi>
Date: Mon, 27 Apr 2009 13:57:29 +0300
Subject: [PATCH] Non-blocking file writer implementation

---
 modules/access_output/file.c |  192 +++++++++++++++++++++++++++++++++++++++---
 1 files changed, 181 insertions(+), 11 deletions(-)

diff --git a/modules/access_output/file.c b/modules/access_output/file.c
index 7639a0b..27720f1 100644
--- a/modules/access_output/file.c
+++ b/modules/access_output/file.c
@@ -64,6 +64,14 @@ static void Close( vlc_object_t * );
 #define APPEND_TEXT N_("Append to file")
 #define APPEND_LONGTEXT N_( "Append to file if it exists instead " \
                             "of replacing it.")
+#define NONBLOCK_TEXT N_("Non-blocking file write")
+#define NONBLOCK_LONGTEXT N_( "Will open the output file in nonblocking mode and grow a buffer if data cannot be written.")
+
+#define NB_BUFFER_TEXT N_("Non-blocking write buffer size")
+#define NB_BUFFER_LONGTEXT N_( "Maximum size for Non-blocking write buffer in kilobytes. 0 = unlimited")
+
+#define OUTPUT_DATA_WAIT_SEC 1
+
 
 vlc_module_begin ()
     set_description( N_("File stream output") )
@@ -75,6 +83,8 @@ vlc_module_begin ()
     add_shortcut( "stream" )
     add_bool( SOUT_CFG_PREFIX "append", 0, NULL, APPEND_TEXT,APPEND_LONGTEXT,
               true )
+    add_bool( SOUT_CFG_PREFIX "non-block", false, NULL, NONBLOCK_TEXT,NONBLOCK_LONGTEXT, false);
+    add_integer( SOUT_CFG_PREFIX "non-block-bufsize", 1024, NULL, NB_BUFFER_TEXT, NB_BUFFER_LONGTEXT, true );
     set_callbacks( Open, Close )
 vlc_module_end ()
 
@@ -83,17 +93,37 @@ vlc_module_end ()
  * Exported prototypes
  *****************************************************************************/
 static const char *const ppsz_sout_options[] = {
-    "append", NULL
+    "append", 
+    "non-block",
+    "non-block-bufsize",
+    NULL
 };
 
+static void* ThreadWrite( vlc_object_t * );
+static ssize_t WriteNonBlock( sout_access_out_t *, block_t * );
 static ssize_t Write( sout_access_out_t *, block_t * );
 static int Seek ( sout_access_out_t *, off_t  );
 static ssize_t Read ( sout_access_out_t *, block_t * );
 static int Control( sout_access_out_t *, int, va_list );
 
+
+typedef struct sout_access_thread_t
+{
+    VLC_COMMON_MEMBERS
+
+    block_fifo_t *p_fifo;
+    int         i_handle;
+
+    int64_t i_max_buf;
+    int discard_count;
+
+} sout_access_thread_t;
+
+
 struct sout_access_out_sys_t
 {
     int i_handle;
+    sout_access_thread_t *p_thread;
 };
 
 /*****************************************************************************
@@ -102,6 +132,7 @@ struct sout_access_out_sys_t
 static int Open( vlc_object_t *p_this )
 {
     sout_access_out_t   *p_access = (sout_access_out_t*)p_this;
+    sout_access_out_sys_t *p_sys;
     int                 fd;
 
     config_ChainParse( p_access, SOUT_CFG_PREFIX, ppsz_sout_options, p_access->p_cfg );
@@ -113,6 +144,33 @@ static int Open( vlc_object_t *p_this )
     }
 
     bool append = var_GetBool( p_access, SOUT_CFG_PREFIX "append" );
+    bool nonblock = var_GetBool( p_access, SOUT_CFG_PREFIX "non-block" );
+
+    if( !( p_sys = calloc ( 1, sizeof( sout_access_out_sys_t ) ) ) )
+        return VLC_ENOMEM;
+    p_access->p_sys = p_sys;
+
+
+    p_sys->p_thread = vlc_object_create( p_access, sizeof( sout_access_thread_t ) );
+
+    if( !p_sys->p_thread )
+    {
+        free (p_sys);
+        return VLC_ENOMEM;
+    }
+ 
+    vlc_object_attach( p_sys->p_thread, p_access );     
+    p_sys->p_thread->p_fifo = block_FifoNew();
+    p_sys->p_thread->i_max_buf = (int64_t)1024 * var_GetInteger( p_access, SOUT_CFG_PREFIX "non-block-bufsize");
+    p_sys->p_thread->discard_count = 0;
+    
+    if( !p_sys->p_thread->p_fifo )
+    {
+        msg_Err( p_access, "error while reserving fifo buffer" );
+        vlc_object_release( p_sys->p_thread );
+        free (p_sys);
+        return VLC_ENOMEM;
+    }
 
     if( !strcmp( p_access->psz_path, "-" ) )
     {
@@ -133,6 +191,7 @@ static int Open( vlc_object_t *p_this )
         path_sanitize( psz_tmp );
 
         fd = utf8_open( psz_tmp, O_RDWR | O_CREAT | O_LARGEFILE |
+                        (nonblock ? O_NONBLOCK : 0) |
                         (append ? 0 : O_TRUNC), 0666 );
         free( psz_tmp );
     }
@@ -140,16 +199,29 @@ static int Open( vlc_object_t *p_this )
     if (fd == -1)
     {
         msg_Err( p_access, "cannot open `%s' (%m)", p_access->psz_path );
+        block_FifoRelease(p_sys->p_thread->p_fifo);
+        free (p_sys);
         return VLC_EGENERIC;
     }
 
-    p_access->pf_write = Write;
+    p_access->pf_write = nonblock ? WriteNonBlock : Write;
     p_access->pf_read  = Read;
     p_access->pf_seek  = Seek;
     p_access->pf_control = Control;
-    p_access->p_sys    = (void *)(intptr_t)fd;
+    p_sys->p_thread->i_handle = fd;
 
-    msg_Dbg( p_access, "file access output opened (%s)", p_access->psz_path );
+    if ( vlc_thread_create( p_sys->p_thread, "sout write thread", ThreadWrite,
+                           VLC_THREAD_PRIORITY_OUTPUT) )
+    {
+        msg_Err( p_access, "cannot spawn sout access thread" );
+        block_FifoRelease(p_sys->p_thread->p_fifo);
+        close (fd);
+        vlc_object_release( p_sys->p_thread );
+        free (p_sys);
+        return VLC_EGENERIC;
+    }
+
+    msg_Dbg( p_access, "file access output opened (%s)%s", p_access->psz_path , nonblock ? " in nonblocking mode" : "");
     if (append)
         lseek (fd, 0, SEEK_END);
 
@@ -157,15 +229,78 @@ static int Open( vlc_object_t *p_this )
 }
 
 /*****************************************************************************
+ * ThreadWrite: Write a packet to the file when it is possible
+ *****************************************************************************/
+static void* ThreadWrite( vlc_object_t *p_this )
+{
+    sout_access_thread_t *p_thread = (sout_access_thread_t*)p_this;
+    while( vlc_object_alive (p_thread) )
+    {
+        block_t *p_buffer;	
+
+        // this will block until data available
+        p_buffer = block_FifoGet( p_thread->p_fifo );
+
+        if( p_buffer == NULL )
+            continue; /* forced wake-up */		
+        
+        while (vlc_object_alive (p_thread))
+            {
+                fd_set rfds;
+                struct timeval tv;
+
+                // Wait for output to be ready
+                FD_ZERO(&rfds);
+                FD_SET(p_thread->i_handle, &rfds);
+
+                tv.tv_sec = OUTPUT_DATA_WAIT_SEC;
+                tv.tv_usec = 0;
+
+                if (select(p_thread->i_handle+1, NULL, &rfds, NULL, &tv) < 1)
+                    continue;
+
+                ssize_t val = write (p_thread->i_handle, p_buffer->p_buffer, p_buffer->i_buffer);
+
+                if (val == -1)
+                    continue;
+
+                if ((size_t)val < p_buffer->i_buffer)
+                {
+                    p_buffer->p_buffer += val;
+                    p_buffer->i_buffer -= val;
+                    continue;
+                }
+
+                // now we know it has been fully written, release the memory
+                block_ChainRelease (p_buffer); 
+                break;
+	        }
+		
+		
+	}
+	return NULL;
+}
+
+
+/*****************************************************************************
  * Close: close the target
  *****************************************************************************/
 static void Close( vlc_object_t * p_this )
 {
     sout_access_out_t *p_access = (sout_access_out_t*)p_this;
+    sout_access_out_sys_t *p_sys = p_access->p_sys;
+    vlc_object_kill( p_sys->p_thread );
+    block_FifoWake( p_sys->p_thread->p_fifo );
+    vlc_thread_join( p_sys->p_thread );
 
-    close( (intptr_t)p_access->p_sys );
+    close( p_sys->p_thread->i_handle );
 
     msg_Dbg( p_access, "file access output closed" );
+    block_FifoRelease(p_sys->p_thread->p_fifo);
+    vlc_object_detach( p_sys->p_thread );
+    vlc_object_release( p_sys->p_thread );
+    free( p_sys );
+
 }
 
 static int Control( sout_access_out_t *p_access, int i_query, va_list args )
@@ -190,12 +325,15 @@ static int Control( sout_access_out_t *p_access, int i_query, va_list args )
  *****************************************************************************/
 static ssize_t Read( sout_access_out_t *p_access, block_t *p_buffer )
 {
+    sout_access_out_sys_t *p_sys = p_access->p_sys;
     ssize_t val;
 
-    do
-        val = read( (intptr_t)p_access->p_sys, p_buffer->p_buffer,
-                    p_buffer->i_buffer );
-    while (val == -1 && errno == EINTR);
+    do {
+      
+        val = read( p_sys->p_thread->i_handle, p_buffer->p_buffer,
+              p_buffer->i_buffer );
+    } while (val == -1 && errno == EINTR);
+
     return val;
 }
 
@@ -204,11 +342,12 @@ static ssize_t Read( sout_access_out_t *p_access, block_t *p_buffer )
  *****************************************************************************/
 static ssize_t Write( sout_access_out_t *p_access, block_t *p_buffer )
 {
+    sout_access_out_sys_t *p_sys = p_access->p_sys;
     size_t i_write = 0;
 
     while( p_buffer )
     {
-        ssize_t val = write ((intptr_t)p_access->p_sys,
+        ssize_t val = write (p_sys->p_thread->i_handle,
                              p_buffer->p_buffer, p_buffer->i_buffer);
         if (val == -1)
         {
@@ -235,9 +374,40 @@ static ssize_t Write( sout_access_out_t *p_access, block_t *p_buffer )
 }
 
 /*****************************************************************************
+* WriteNonBlock: NonBlocking buffered write on a file descriptor.
+*****************************************************************************/
+static ssize_t WriteNonBlock( sout_access_out_t *p_access, block_t *p_buffer )
+{
+    sout_access_out_sys_t *p_sys = p_access->p_sys;
+    size_t i_total_size = 0;
+
+    // new block is always put on the fifo
+    i_total_size = block_FifoPut(p_sys->p_thread->p_fifo, p_buffer);
+
+    // destroy items from the buffer until it meets the memory limits
+    if (p_sys->p_thread->i_max_buf > 0)
+        while ( block_FifoSize(p_sys->p_thread->p_fifo) > p_sys->p_thread->i_max_buf)
+	    {
+	        p_buffer = block_FifoGet(p_sys->p_thread->p_fifo);
+            block_Release (p_buffer);	
+
+            if ((p_sys->p_thread->discard_count % 10000) == 0)
+            {
+                msg_Err( p_access, "fifo buffer full, total number of packets discarded = %d ", p_sys->p_thread->discard_count );
+            }
+            p_sys->p_thread->discard_count++;
+        }
+
+    return i_total_size;
+}
+
+
+/*****************************************************************************
  * Seek: seek to a specific location in a file
  *****************************************************************************/
 static int Seek( sout_access_out_t *p_access, off_t i_pos )
 {
-    return lseek( (intptr_t)p_access->p_sys, i_pos, SEEK_SET );
+    sout_access_out_sys_t *p_sys = p_access->p_sys;
+
+    return lseek( p_sys->p_thread->i_handle, i_pos, SEEK_SET );
 }
-- 
1.6.0.4

